// Defines interface to libflowXXX.dll plugins
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#ifndef LIBFLOW_H
#define LIBFLOW_H

typedef void* HFLOWLIB;

#ifdef __FLOWLIB_INTERNAL__
typedef unsigned char BYTE;
#endif

enum FlowLib_Algo
{
    AlgoUnknown=0,

	AlgoFast_1dir =1,//MFlow "1-dir"
	AlgoFast_2dir =2,//MFlow "2-dir"

	AlgoNoMask1 =11,//MBlockFps, mode=0
	AlgoNoMask2 =12,//MBlockFps, mode=1
	AlgoNoMask3 =13,//MBlockFps, mode=2

	AlgoSimple =21, //like FlowInterSimple from MaskFun.cpp OR MBlockFps, mode=3
	AlgoNormal =22, //FlowInter
	AlgoExtended=23, //FlowInterExtra

	AlgoMask =127,
	AlgoProcessBlocks =128
};

#define FLOWLIB_VERSION		0x0002

#define FLOWLIB_TYPE_CPU    0
#define FLOWLIB_TYPE_GPU    0x010000

#define FLOWLIB_FLAG_BLOCK  1
#define FLOWLIB_FLAG_SAD    2

typedef int (*fnFlowLib_GetVersion)();

typedef HFLOWLIB (*fnFlowLib_Init)(int width,int height,
							   int blkX,int blkY,
							   int nBlkX,int nBlkY,
							   int marginX,int marginY,
							   int uvRatioY,int nPel,
							   int algo,int flags);

typedef void (*fnFlowLib_Clear)(HFLOWLIB hFlowLib);

typedef void (*fnFlowLib_SetTime)(HFLOWLIB hFlowLib,int time);

typedef void (*fnFlowLib_Calculate)(HFLOWLIB hFlowLib,int algoType,
							//destination frame for Y-channel
							BYTE *pDst, int dstLine,
							//destination frames for UV-channel
							BYTE *uvDst1, BYTE *uvDst2, int uvDstLine,
							//reference frames for Y-channel
							int nRefBFrame, const BYTE *pRef_B,
							int nRefFFrame, const BYTE *pRef_F, int refLine,
                            //reference frames for UV-channel
							const BYTE *uvRef1_B, const BYTE *uvRef1_F,
							const BYTE *uvRef2_B, const BYTE *uvRef2_F, int uvRefLine,
							//vectors
							const BYTE *VX_B, const BYTE *VX_F, const BYTE *VY_B, const BYTE *VY_F,
							const BYTE *VX_BB, const BYTE *VX_FF, const BYTE *VY_BB, const BYTE *VY_FF,
							//masks
							const BYTE *Mask_B, const BYTE *Mask_F,
							const BYTE *SAD_B, const BYTE *SAD_F);

#endif //LIBFLOW_H
